/*
 * LoraBBS
 * Copyright (c) 1987-98,2012-14 Marco Maccaferri and Others
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EDITOR_H_
#define EDITOR_H_

#include <string>
#include <map>
#include <vector>

#define BLACK           0
#define BLUE            1
#define GREEN           2
#define CYAN            3
#define RED             4
#define MAGENTA         5
#define BROWN           6
#define LGREY           7
#define DGREY           8
#define LBLUE           9
#define LGREEN          10
#define LCYAN           11
#define LRED            12
#define LMAGENTA        13
#define YELLOW          14
#define WHITE           15

#define UP              0x4800
#define DOWN            0x5000
#define LEFT            0x4B00
#define RIGHT           0x4D00
#define PGDN            0x5100
#define PGUP            0x4900
#define HOME            0x4700
#define END             0x4F00
#define RETURN          0x0D
#define DEL             0x7F

#define CTRLA           1   // \001 \x01
#define CTRLB           2   // \002 \x02
#define CTRLC           3   // \003 \x03
#define CTRLD           4   // \004 \x04
#define CTRLE           5   // \005 \x05
#define CTRLF           6   // \006 \x06
#define CTRLG           7   // \007 \x07
#define CTRLH           8   // \010 \x08
#define CTRLI           9   // \011 \x09
#define CTRLJ           10  // \012 \x0A
#define CTRLK           11  // \013 \x0B
#define CTRLL           12  // \014 \x0C
#define CTRLM           13  // \015 \x0D
#define CTRLN           14  // \016 \x0E
#define CTRLO           15  // \017 \x0F
#define CTRLP           16  // \020 \x10
#define CTRLQ           17  // \021 \x11
#define CTRLR           18  // \022 \x12
#define CTRLS           19  // \023 \x13
#define CTRLT           20  // \024 \x14
#define CTRLU           21  // \025 \x15
#define CTRLV           22  // \026 \x16
#define CTRLW           23  // \027 \x17
#define CTRLX           24  // \030 \x18
#define CTRLY           25  // \031 \x19
#define CTRLZ           26  // \032 \x2A

namespace bbs {

    class bufferedio {
    public:
        bufferedio();
        bufferedio(int fd_io);
        bufferedio(int fd_in, int fd_out);
        virtual ~bufferedio();

        int fd_in;
        int fd_out;

        int rx_size;
        int rx_position;
        unsigned char rx_buffer[512];

        int tx_position;
        unsigned char tx_buffer[2048];

        int read(int tout = 0);

        void write(unsigned char c);
        void write(const unsigned char * bytes, int size);

        void write(const char * str);

        void clear();
        void flush();

        void fill_rx_buffer(int tout = 0);
    };

    #define TERMINAL_MODE_TTY    0
    #define TERMINAL_MODE_ANSI   1
    #define TERMINAL_MODE_AVATAR 2

    class terminal: public bufferedio {
    public:
        terminal(int fd_in, int fd_out, int width = 80, int height = 24);
        virtual ~terminal();

        int width;
        int height;
        int mode;

        int col;
        int row;
        int foreground;
        int background;
        int highlight;
        int tab_size;

        int pending_foreground;
        int pending_background;

        char printf_temp[2048];

        void cls();
        void clreol();
        void gotoxy(int x, int y);

        int get_column();
        int get_row();

        void set_background(int c);
        void set_foreground(int c);

        int get_char(int tout = 0);
        std::string get_string(int width = -1, int echo = -1);

        void put_char(int c);
        void put_string(const char * s);
        void put_string(const char * s, int len);
        void put_string(const std::string& s);
        void put_string(const std::string& s, int len);

        void printf(const char * fmt, ...);

        void clear();
        void flush();

        int get_width();
        int get_height();

        void flush_pending_attributes();
    };

    class editor {
    public:
        editor(terminal *term);
        virtual ~editor();

        int foreground;
        int background;
        int quote_foreground;
        int toolbar_foreground;
        int toolbar_background;

        int window_top;
        int window_height;
        int toolbar_row;

        std::string text;

        int run();

        int get_row_at(int index);
        int get_col_at(int index);
        int get_index(int row, int col);

        void go_to_index(int index);

        terminal * term;

        std::map<int,int> row_index;
        std::map<int,int> row_size;
        int row;
        int col;
        int index;
        int wrap_column;
        int top_row;

        void initalize();
        void redraw_screen(int from_row, bool force);
        void process_char(int c);
        void gotoxy(int x, int y);
    };

} /* namespace bbs */

#endif /* EDITOR_H_ */
