/*
 * LoraBBS
 * Copyright (c) 1987-98,2012-14 Marco Maccaferri and Others
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <libintl.h>
#include <sys/stat.h>

#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <string>

#include "editor.h"

using namespace std;
using namespace bbs;

static int tty_fd_in;
static int tty_fd_out;
static struct termios old_termio;
static struct vt_mode old_vtm;

static void initialize_terminal() {
    int value;
    socklen_t value_len;
    struct termios new_termio;
    struct vt_mode new_vtm;

    value = 0;
    value_len = sizeof(value);
    if (getsockopt(tty_fd_out, SOL_SOCKET, SO_TYPE, &value, &value_len) == 0) {
        // IAC WILL ECHO IAC WILL SUPPRESS_GO_AHEAD IAC WONT LINEMODE
        // 255  251    1 255  251                 3 255  252       34
        write(tty_fd_out, "\377\373\001\377\373\003\377\374\042", 9);
    }

    value = 0;
    value_len = sizeof(value);
    if (getsockopt(tty_fd_in, SOL_SOCKET, SO_TYPE, &value, &value_len) == 0) {
        fd_set fds;
        struct timeval tv;
        unsigned char rx_buffer[16];

        for (;;) {
            FD_ZERO(&fds);
            FD_SET(tty_fd_in, &fds);

            tv.tv_sec = 1;
            tv.tv_usec = 0;

            int rc = select(tty_fd_in + 1, &fds, NULL, NULL, &tv);
            if (rc <= 0) {
                break;
            }

            if (FD_ISSET(tty_fd_in, &fds)) {
                read(tty_fd_in, rx_buffer, sizeof(rx_buffer));
            }
        }
    }
    else {
        fcntl(tty_fd_in, F_SETFL, O_NONBLOCK);

        ioctl(tty_fd_in, TCGETS, &old_termio);
        new_termio = old_termio;
        new_termio.c_iflag &= ~(IGNBRK | BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON);
        new_termio.c_oflag = 0;
        new_termio.c_lflag &= ~(ECHO | ECHONL | ICANON | IEXTEN | ISIG);
        new_termio.c_cflag &= ~(CSIZE | PARENB);
        new_termio.c_cflag |= CS8;
        new_termio.c_cc[VMIN] = 1;
        new_termio.c_cc[VTIME] = 0;
        ioctl(tty_fd_in, TCSETSW, &new_termio);

        if (isatty(tty_fd_in)) {
            ioctl(tty_fd_in, VT_GETMODE, &old_vtm);
            ioctl(tty_fd_in, KDSKBMODE, K_RAW);
            new_vtm = old_vtm;
            new_vtm.mode = VT_PROCESS;
            new_vtm.relsig = SIGUSR1;
            new_vtm.acqsig = SIGUSR2;
            ioctl(tty_fd_in, VT_SETMODE, &new_vtm);
        }
    }
}

static void restore_terminal() {
    int value = 0;
    socklen_t value_len = sizeof(value);

    if (getsockopt(tty_fd_in, SOL_SOCKET, SO_TYPE, &value, &value_len) == 0) {
        return;
    }

    if (tty_fd_in != -1) {
        ioctl(tty_fd_in, TCSETSW, &old_termio);
        if (isatty(tty_fd_in)) {
            ioctl(tty_fd_in, KDSKBMODE, K_XLATE);
            ioctl(tty_fd_in, VT_SETMODE, &old_vtm);
        }
    }
}

static void signal_callback_handler(int signum) {
    restore_terminal();
    exit(0);
}

int main(int argc, char * argv[]) {

    tty_fd_in = fileno(stdin);
    tty_fd_out = fileno(stdout);

    initialize_terminal();

    signal(SIGABRT, signal_callback_handler);
    signal(SIGFPE, signal_callback_handler);
    signal(SIGSEGV, signal_callback_handler);
    signal(SIGKILL, signal_callback_handler);
    signal(SIGHUP, signal_callback_handler);
    signal(SIGTERM, signal_callback_handler);
    signal(SIGINT, signal_callback_handler);

    terminal * term = new terminal(tty_fd_in, tty_fd_out);

    editor * app = new editor(term);
    app->run();

    term->flush();

    delete app;
    delete term;

    restore_terminal();

    return 0;
}
