/*
 * LoraBBS
 * Copyright (c) 1987-98,2012-14 Marco Maccaferri and Others
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>

#include <string>
#include <sstream>

#include "editor.h"

using namespace std;

#define DEFAULT         -1

namespace bbs {

    bufferedio::bufferedio() {
        this->fd_in = fileno(stdin);
        this->fd_out = fileno(stdout);
        this->rx_size = this->rx_position = 0;
        this->tx_position = 0;
    }

    bufferedio::bufferedio(int fd_io) {
        this->fd_in = fd_io;
        this->fd_out = fd_io;
        this->rx_size = this->rx_position = 0;
        this->tx_position = 0;
    }

    bufferedio::bufferedio(int fd_in, int fd_out) {
        this->fd_in = fd_in;
        this->fd_out = fd_out;
        this->rx_size = this->rx_position = 0;
        this->tx_position = 0;
    }

    bufferedio::~bufferedio() {

    }

    int bufferedio::read(int tout /*= 0*/) {
        if (rx_size == 0) {
            fill_rx_buffer(tout);
        }

        if (rx_size != 0) {
            unsigned char c = rx_buffer[rx_position++];
            rx_size--;
            return c;
        }

        return -1;
    }

    void bufferedio::fill_rx_buffer(int tout /*= 0*/) {
        fd_set fds;
        struct timeval tv;

        FD_ZERO(&fds);
        FD_SET(fd_in, &fds);

        tv.tv_sec = tout / 1000;
        tv.tv_usec = (tout % 1000) * 1000;

        int rc = select(fd_in + 1, &fds, NULL, NULL, tout != 0 ? &tv : NULL);
        if (rc < 0 || (tout == 0 && rc == 0)) {
            raise(SIGHUP);
            return;
        }

        if (FD_ISSET(fd_in, &fds)) {
            int rc = ::read(fd_in, rx_buffer, sizeof(rx_buffer));
            if (rc > 0) {
                rx_size = rc;
                rx_position = 0;
                return;
            }
            raise(SIGHUP);
        }
    }

    void bufferedio::write(unsigned char byte) {
        if (tx_position >= sizeof(tx_buffer)) {
            flush();
        }
        tx_buffer[tx_position++] = byte;
    }

    void bufferedio::write(const unsigned char * bytes, int size) {
        while (size > 0) {
            if (tx_position >= sizeof(tx_buffer)) {
                flush();
            }

            int s = sizeof(tx_buffer) - tx_position;
            if (size < s) {
                s = size;
            }
            memcpy(&tx_buffer[tx_position], bytes, s);
            tx_position += s;
            bytes += s;
            size -= s;
        }
    }

    void bufferedio::write(const char * str) {
        int size = strlen(str);

        while (size > 0) {
            if (tx_position >= sizeof(tx_buffer)) {
                flush();
            }

            int s = sizeof(tx_buffer) - tx_position;
            if (size < s) {
                s = size;
            }

            memcpy(&tx_buffer[tx_position], str, s);
            tx_position += s;

            str += s;
            size -= s;
        }
    }

    void bufferedio::clear() {
        tx_position = 0;
    }

    void bufferedio::flush() {
        unsigned char *p = tx_buffer;

        while (tx_position > 0) {
            int i = ::write(fd_out, p, tx_position);
            if (i == -1) {
                raise(SIGHUP);
                return;
            }
            if (i > 0) {
                tx_position -= i;
                p += i;
            }
        }
    }

    terminal::terminal(int fd_in, int fd_out, int width /*= 80*/, int height /*= 24*/) : bufferedio(fd_in, fd_out) {
        this->width = width;
        this->height = height;

        this->tab_size = 8;
        this->mode = TERMINAL_MODE_ANSI;

        this->col = 0;
        this->row = 0;
        this->foreground = this->pending_foreground = DEFAULT;
        this->background = this->pending_background = DEFAULT;
        this->highlight = DEFAULT;
    }

    terminal::~terminal() {

    }

    void terminal::cls() {
        flush_pending_attributes();
        if (mode == TERMINAL_MODE_ANSI) {
            write("\x1B[2J\x1B[1;1f");
        }
        else {
            write(CTRLL);
        }
        col = 0;
        row = 0;
    }

    void terminal::clreol() {
        flush_pending_attributes();
        if (mode == TERMINAL_MODE_ANSI) {
            write("\x1B[K");
        }
        else if (mode == TERMINAL_MODE_AVATAR) {
            write("\x16\x07");
        }
    }

    void terminal::gotoxy(int x, int y) {
        char temp[20];

        x %= width;
        y %= height;

        if (mode == TERMINAL_MODE_ANSI) {
            sprintf(temp, "\x1B[%d;%df", y + 1, x + 1);
        }
        else {
            sprintf(temp, "\x16\x08%c%c", y + 1, x + 1);
        }
        write(temp);

        col = x;
        row = y;
    }

    int terminal::get_column() {
        return col;
    }

    int terminal::get_row() {
        return row;
    }

    void terminal::set_background(int c) {
        pending_background = c;
    }

    void terminal::set_foreground(int c) {
        pending_foreground = c;
    }

    void terminal::put_char(int c) {
        flush_pending_attributes();

        if (c == 0x0C) {
            cls();
            return;
        }

        write((unsigned char )c);

        switch (c) {
            case 0x07:
                // Bell, do nothing
                break;
            case 0x08:
                if (col > 0) {
                    col--;
                }
                break;
            case 0x09: // TAB
                col = ((col / tab_size) + 1) * tab_size;
                if (col > width) {
                    col -= width;
                    if (row < (height - 1)) {
                        row++;
                    }
                }
                break;
            case 0x0A: // LF
                if (row < (height - 1)) {
                    row++;
                }
                break;
            case 0x0D: // CR
                col = 0;
                break;
            default:
                if (col >= width) {
                    col = 0;
                    if (row < (height - 1)) {
                        row++;
                    }
                }
                col++;
                break;
        }
    }

    void terminal::put_string(const char * s) {
        while (*s) {
            put_char(*s++);
        }
    }

    void terminal::put_string(const char * s, int len) {
        for (int i = 0; i < len && *s; i++) {
            put_char(*s++);
        }
    }

    void terminal::put_string(const string& s) {
        const char * p = s.c_str();
        while (*p) {
            put_char(*p++);
        }
    }

    void terminal::put_string(const string& s, int len) {
        const char * p = s.c_str();
        for (int i = 0; i < len && *p; i++) {
            put_char(*p++);
        }
    }

    void terminal::printf(const char * fmt, ...) {
        va_list va;

        va_start(va, fmt);
        vsnprintf(printf_temp, sizeof(printf_temp), fmt, va);
        va_end(va);

        char * p = printf_temp;
        while (*p) {
            put_char(*p++);
        }
    }

    void terminal::flush_pending_attributes() {
        if (pending_foreground == foreground && pending_background == background) {
            return;
        }

        if (mode == TERMINAL_MODE_ANSI) {
            string s = "\x1B[";

            if (pending_foreground != foreground) {
                int hl = highlight;

                switch (pending_foreground) {
                    case BLACK:
                        case BLUE:
                        case GREEN:
                        case CYAN:
                        case RED:
                        case MAGENTA:
                        case BROWN:
                        case LGREY:
                        hl = 0;
                        break;
                    case DGREY:
                        case LBLUE:
                        case LGREEN:
                        case LCYAN:
                        case LRED:
                        case LMAGENTA:
                        case YELLOW:
                        case WHITE:
                        hl = 1;
                        break;
                }

                if (hl != highlight) {
                    s += (hl ? "1" : "0");
                    s += ";";
                    if (!hl) {
                        background = DEFAULT;
                    }
                    highlight = hl;
                }

                switch (pending_foreground) {
                    case BLACK:
                    case DGREY:
                        s += "30";
                        break;
                    case BLUE:
                    case LBLUE:
                        s += "34";
                        break;
                    case GREEN:
                    case LGREEN:
                        s += "32";
                        break;
                    case CYAN:
                    case LCYAN:
                        s += "36";
                        break;
                    case RED:
                    case LRED:
                        s += "31";
                        break;
                    case MAGENTA:
                    case LMAGENTA:
                        s += "35";
                        break;
                    case BROWN:
                    case YELLOW:
                        s += "33";
                        break;
                    case LGREY:
                    case WHITE:
                        s += "37";
                        break;
                }

                foreground = pending_foreground;
            }

            if (pending_background != background) {
                if (s.length() != 2) {
                    s += ";";
                }
                switch (pending_background) {
                    case BLACK:
                        s += "40";
                        break;
                    case BLUE:
                        s += "44";
                        break;
                    case GREEN:
                        s += "42";
                        break;
                    case CYAN:
                        s += "46";
                        break;
                    case RED:
                        s += "41";
                        break;
                    case MAGENTA:
                        s += "45";
                        break;
                    case BROWN:
                        s += "43";
                        break;
                    case LGREY:
                        s += "47";
                        break;
                }
                background = pending_background;
            }

            if (s.length() != 2) {
                s += "m";
                write(s.c_str());
            }
        }
        else if (mode == TERMINAL_MODE_AVATAR) {
            write(CTRLV);
            write(CTRLA);
            write(pending_foreground | (pending_background << 4));
            foreground = pending_foreground;
            background = pending_background;
        }
    }

    int terminal::get_char(int tout /*= 0*/) {
        int state, c, pending;

        flush();

        state = 0;

        while ((c = read(tout)) != -1) {
            switch (state) {
                case 0:
                    if (c == 0x1B) {
                        state = 1;
                        break;
                    }
                    else if (c == CTRLV) {
                        state = 5;
                        break;
                    }
                    return c;
                case 1:
                    if (c == '[') {
                        state = 2;
                        break;
                        return c;
                    }
                    if (c == 'O') {
                        state = 3;
                        break;
                    }
                    state = 0;
                    break;
                case 2:
                    if (c == '3') {
                        pending = DEL;
                        state = 4;
                    }
                    if (c == '5') {
                        pending = PGUP;
                        state = 4;
                    }
                    if (c == '6') {
                        pending = PGDN;
                        state = 4;
                    }
                    if (c == 'A') {
                        return UP;
                    }
                    if (c == 'B') {
                        return DOWN;
                    }
                    if (c == 'C') {
                        return RIGHT;
                    }
                    if (c == 'D') {
                        return LEFT;
                    }
                    if (c == 'H') {
                        return HOME;
                    }
                    if (c == 'K') {
                        return END;
                    }
                    if (!isdigit(c)) {
                        state = 0;
                    }
                    break;
                case 3:
                    if (c == 'F') {
                        return END;
                    }
                    if (c == 'H') {
                        return HOME;
                    }
                    state = 0;
                    break;
                case 4:
                    if (c == 0x7E) {
                        return pending;
                    }
                    state = 0;
                    break;
                case 5:
                    if (c == CTRLC) {
                        return UP;
                    }
                    if (c == CTRLD) {
                        return DOWN;
                    }
                    if (c == CTRLE) {
                        return LEFT;
                    }
                    if (c == CTRLF) {
                        return RIGHT;
                    }
                    state = 0;
                    break;
            }
        }

        return -1;
    }

    string terminal::get_string(int width /*= -1*/, int echo /*= -1*/) {
        int c;
        string s;

        flush_pending_attributes();

        if (width != -1 && s.length() > width) {
            s = s.substr(0, width);
        }
        if (echo == -1) {
            put_string(s);
        }
        else {
            for (int i = 0; i < s.length(); i++) {
                put_char((char) echo);
            }
        }

        if (width != -1) {
            for (int i = s.length(); i < width; i++) {
                put_char(' ');
            }
            for (int i = s.length(); i < width; i++) {
                put_char((char) 0x08);
            }
        }

        flush();

        while ((c = read(0)) != 0x0D && c != 0x0A) {
            switch (c) {
                case 0x08:
                case 0x7F:
                    if (s.length() > 0) {
                        s = s.substr(0, s.length() - 1);
                        put_string("\010 \010");
                    }
                    break;
                default:
                    if (c < 0x20 || c >= 0x7F) {
                        break;
                    }
                    if (width == -1 || s.length() < width) {
                        s += (char) c;
                        put_char(echo == -1 ? (char) c : (char) echo);
                    }
                    break;
            }
            bufferedio::flush();
        }

        return s;
    }

    void terminal::clear() {
        pending_foreground = foreground;
        pending_background = background;
        bufferedio::clear();
    }

    void terminal::flush() {
        flush_pending_attributes();
        bufferedio::flush();
    }

    int terminal::get_width() {
        return width;
    }

    int terminal::get_height() {
        return height;
    }

} /* namespace bbs */
